# Plugin for Foswiki - The Free and Open Source Wiki, http://foswiki.org/
#
# ICalPlugin is Copyright (C) 2011-2020 Michael Daum http://michaeldaumconsulting.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details, published at
# http://www.gnu.org/copyleft/gpl.html

package Foswiki::Plugins::ICalPlugin::Core;

use strict;
use warnings;

use Foswiki::Func ();
use Foswiki::Sandbox ();
use Foswiki::Form ();
use Foswiki::Plugins ();
use Foswiki::Plugins::ICalPlugin::Calendar ();
use DateTime::Duration ();
use DateTime::Format::ICal ();
use DateTime::Span ();
use Data::ICal::Entry::Event ();
use Error qw(:try);
use Encode ();

use constant TRACE => 0;    # toggle me
#use Data::Dump qw(dump);

###############################################################################
sub new {
  my $class = shift;

  my $this = bless({@_}, $class);

  return $this;
}

###############################################################################
sub FORMATICAL {
  my ($this, $session, $params, $theTopic, $theWeb) = @_;

  #_writeDebug("### called FORMATICAL()");

  my $web = $params->{web} || $session->{webName};
  my $topic = $params->{topic} || $session->{topicName};
  my $attachment = $params->{attachment} || 'calendar.ics';
  my $text = $params->{_DEFAULT} || $params->{text};
  my $url = $params->{url};
  my $skip = $params->{skip} || 0;
  my $limit = $params->{limit};

  ($web, $topic) = Foswiki::Func::normalizeWebTopicName($web, $topic);

  return _inlineError("topic $web.$topic does not exist")
    unless Foswiki::Func::topicExists($web, $topic);

  my $error;
  my $request = Foswiki::Func::getRequestObject();
  my $expire = $params->{expire};
  my $doRefresh = $request->param("refresh") || '';

  my $cal = Foswiki::Plugins::ICalPlugin::Calendar->new();
  $cal->clearCache() if $doRefresh =~ /^(on|ical)$/;

  if (defined $url) {
    # get calendar from url
    try {
      $cal->getDataFromUrl($url, $expire);
    }
    catch Error::Simple with {
      $error = "ERROR fetching data from url: " . shift->{-text};
    }
  } elsif (defined $text) {
    # get calendar from inline text
    $cal->parseData($text, undef, $expire);
  } else {
    # get calendar from attachment
    return _inlineError("calendar $attachment not found at $web.$topic")
      unless Foswiki::Func::attachmentExists($web, $topic, $attachment);

    try {
      $text = Foswiki::Func::readAttachment($web, $topic, $attachment);
      $text = Encode::decode_utf8($text);
    }
    catch Foswiki::AccessControlException with {
      $error = "ERROR reading data from attachment: " . shift->{-text};
    };

    unless (defined $error) {
      $cal->parseData($text, "$web.$topic.$attachment", $expire);
    }
  }
  return _inlineError($error) if defined $error;
  return '' unless $cal;

  return $cal->ical->as_string if Foswiki::Func::isTrue($params->{raw});

  my $querySpan = $this->getQuerySpan($params);
  my @events = $cal->getEvents($querySpan);
  _writeDebug("### parsed " . scalar(@events) . " events");

  # sort events
  my $sortCrit;
  my $theSort = $params->{sort} || 'off';
  if ($theSort eq 'start' || $theSort eq 'on') {
    $sortCrit = 'start';
  } elsif ($theSort eq 'end') {
    $sortCrit = 'end';
    #  } elsif ($theSort eq 'modified') {
    #    $sortCrit = '_modifiedepoch';
    #  } elsif ($theSort eq 'created') {
    #    $sortCrit = '_createdepoch';
  } else {
    $theSort = 'off';
  }
  unless ($theSort eq 'off') {
    @events = sort { ($a->{$sortCrit} || 0) <=> ($b->{$sortCrit} || 0) } @events;
  }

  my @result = ();
  my $index = 0;
  #print STDERR "querySpan: start=".$querySpan->start.", end=".$querySpan->end."\n";
  foreach my $event (@events) {
    $index++;

    _writeDebug("### $index: summary=$event->{summary}, start=" . $event->start . ", end=" . $event->end) if TRACE;
    next unless $index > $skip;
    last if $limit && $index > $skip + $limit;

    my $line = $event->format($params);
    #print STDERR "line=$line\n";

    # add params
    while (my ($key, $val) = each %$params) {
      next if $key =~ /^(_.*|start|end)$/;
      next unless defined $val;
      $line =~ s/\$$key/$val/g;
    }

    # clean up
    $line =~ s/\$index/$index/g;
    $line =~ s/\$(?:location|dtstart|start|dtend|until|end|last-modified|modified|description|attendee|organizer|rrule|rdate|summary)//g;
    $line =~ s/\$plainify\(.*?\)//g;
    $line =~ s/\$[mec]?(seco?n?d?s?|minu?t?e?s?|hour?s?|day|w(eek|day)|dow|mo(?:nt?h?)?|ye(?:ar)?)//g;

    push @result, $line;
  }
  _writeDebug("### extracted " . scalar(@result) . " events");

  return '' unless @result;

  my $theHeader = $params->{header} || '';
  my $theFooter = $params->{footer} || '';
  my $theSeparator = $params->{separator};
  $theSeparator = '$n' unless defined $theSeparator;

  my $result = $theHeader . join($theSeparator, @result) . $theFooter;

  $result =~ s/\$n/\n/g;
  $result =~ s/\$perce?nt/\%/g;
  $result =~ s/\$dollar/\$/g;

  return $result;
}

###############################################################################
sub parseTime {
  my $time = shift;

  if (ref($time) eq 'DateTime') {
    return $time->epoch();
  }

  my $epoch;

  $time =~ s/^\s*//;
  $time =~ s/\s*$//;

  # deal with 20111224
  if ($time =~ /^(\d\d\d\d)(\d\d)(\d\d)$/) {
    $time = "$1-$2-$3";
  }

  # deal with 20111224T1200000
  elsif ($time =~ /^(\d\d\d\d)(\d\d)(\d\d)T(\d\d)(\d\d)(\d\d)(Z.*?)$/) {
    $time = "$1-$2-$3T$4:$5:$6$7";
  }

  my $result = Foswiki::Time::parseTime($time);

  unless (defined $result) {
    Foswiki::Func::writeWarning("ICalPlugin::Core - cant parse time '$time'");
    $result = 0;
  }

  return $result;
}

###############################################################################
sub getQuerySpan {
  my ($this, $params) = @_;

  my $theStart = $params->{start};
  my $dtStart = DateTime->from_epoch(epoch => (defined $theStart) ? parseTime($theStart) : time);

  my $theSpan = $params->{span} || '1 month';
  my $theEnd = $params->{end};

  my $dtEnd;
  if (defined $theEnd) {
    $dtEnd = DateTime->from_epoch(epoch => parseTime($theEnd));
  } else {
    my $years = 0;
    my $months = 0;
    my $weeks = 0;
    my $days = 0;
    my $hours = 0;
    my $minutes = 0;
    my $seconds = 0;
    $years = $1 if $theSpan =~ /(\d+)\s+years(s)?/;
    $months = $1 if $theSpan =~ /(\d+)\s+month(s)?/;
    $weeks = $1 if $theSpan =~ /(\d+)\s+week(s)?/;
    $days = $1 if $theSpan =~ /(\d+)\s+day(s)?/;
    $hours = $1 if $theSpan =~ /(\d+)\s+hour(s)?/;
    $minutes = $1 if $theSpan =~ /(\d+)\s+minute(s)?/;
    $seconds = $1 if $theSpan =~ /(\d+)\s+sec(ond)?(s)?/;
    $dtEnd = $dtStart->clone();
    $dtEnd->add(
      years => $years,
      months => $months,
      weeks => $weeks,
      days => $days,
      hours => $hours,
      minutes => $minutes,
      seconds => $seconds,
    );
  }

  #_writeDebug("query span: start=".$dtStart.", end=".$dtEnd) if TRACE;

  return DateTime::Span->from_datetimes(
    start => $dtStart,
    end => $dtEnd
  );
}

###############################################################################
sub afterSaveHandler {
  my ($this, $text, $topic, $web, $meta) = @_;

  unless ($meta) {
    ($meta) = Foswiki::Func::readTopic($web, $topic);
  }

  _writeDebug("called afterSaveHandler");

  my @events = ();
  my $event = $this->getEventFromDataForm($meta);
  push @events, $event if defined $event;

  if (Foswiki::Func::getContext()->{MetaDataPluginEnabled}) {
    push @events, $this->getEventsFromMetaData($meta);
  }

  if (@events) {
    try {
      $this->updateCalendar(\@events);
    }
    catch Error::Simple with {
      my $error = shift;
      print STDERR "ERROR: can't update calendar - " . $error . "\n";
    }
  } else {
    _writeDebug("nothing to update");
  }
}

###############################################################################
sub updateCalendar {
  my ($this, $newEvents, $deleteEvents) = @_;

  _writeDebug("called updateCalendar");

  my ($calendarWeb, $calendarTopic, $calendarAttachment) = $this->getCalendarAttachment;
  return unless defined $calendarWeb;

  my %knownUID;

  # add deletable uids as "known" so they are skipped constructing the updated ical file
  if (defined $deleteEvents) {
    foreach my $event (@$deleteEvents) {
      my $uid = $event->property("uid")->[0]->value;
      $knownUID{$uid} = 1;
      #print STDERR "deleting uid=$uid\n";
    }
  }

  if (defined $newEvents) {
    foreach my $event (@$newEvents) {
      my $uid = $event->property("uid")->[0]->value;
      $knownUID{$uid} = 1;
      #print STDERR "uid=$uid is known\n";
    }
  }

  my $newCal = Data::ICal->new(
    data => "BEGIN:VCALENDAR
PRODID:-//Foswiki///ICalPlugin//$Foswiki::Plugins::ICalPlugin::RELEASE
VERSION:2.0
END:VCALENDAR
"
  );

  if (Foswiki::Func::attachmentExists($calendarWeb, $calendarTopic, $calendarAttachment)) {
    my $data = Foswiki::Func::readAttachment($calendarWeb, $calendarTopic, $calendarAttachment);
    if ($data) {
      $data = Foswiki::Sandbox::untaintUnchecked($data);    # SMELL: some libs barf on tainted strings

      my $oldCal = Data::ICal->new(data => $data);          # SMELL: do we still need this wrapper?

      if (ref($oldCal) eq 'Class::ReturnValue') {
        print STDERR "ERROR:" . $oldCal->error_message . "\n";    # ERROR ... but keep on going
      } else {
        foreach my $entry (@{$oldCal->entries()}) {
          next unless $entry->ical_entry_type() eq 'VEVENT';
          my $uid = $entry->property("uid")->[0]->value;
          unless ($knownUID{$uid}) {
            $newCal->add_entry($entry);
          }
        }
      }
    }
  }

  if (defined $newEvents) {
    foreach my $event (@$newEvents) {
      $newCal->add_entry($event);
    }
  }

  my $data = $newCal->as_string;
  open(my $fh, '<', \$data) or die $!;

  Foswiki::Func::saveAttachment(
    $calendarWeb,
    $calendarTopic,
    $calendarAttachment,
    {
      filesize => length($data),
      stream => $fh
        #SMELL: attachment attributes are lost when saving again (comments, hide, ...)
    }
  );

  close $fh;

  #print STDERR "cal=".dump($newCal)."\n";
  return 1;    # success
}

###############################################################################
sub getCalendarAttachment {
  my $this = shift;

  unless ($this->{calendarWeb}) {
    my $calendar = Foswiki::Func::getPreferencesValue('ICALPLUGIN_CALENDAR') || $Foswiki::cfg{ICalPlugin}{DefaultCalendar};
    $calendar = Foswiki::Func::expandCommonVariables($calendar) if $calendar =~ /%[A-Z]+{.+?}%/;

    #throw Error::Simple("ICALPLUGIN_CALENDAR not defined") unless $calendar;
    return unless $calendar;

    if ($calendar =~ /^($Foswiki::regex{webNameRegex})\.($Foswiki::regex{topicNameRegex})\.(.*?)$/) {
      $this->{calendarWeb} = $1;
      $this->{calendarTopic} = $2;
      $this->{calendarAttachment} = $3;
    } else {
      throw Error::Simple("invalid format of ICALPLUGIN_CALENDAR: $calendar, must be web.topic.attachment");
    }
  }

  _writeDebug("using calendar at web=$this->{calendarWeb} topic=$this->{calendarTopic}, attachment=$this->{calendarAttachment}");

  return ($this->{calendarWeb}, $this->{calendarTopic}, $this->{calendarAttachment});
}

###############################################################################
sub getEventFromDataForm {
  my ($this, $meta) = @_;

  #print STDERR "called getEventFromDataForm\n";

  my $topicType = $this->getField($meta, 'TopicType');
  return unless defined $topicType;

  #print STDERR "topicType=$topicType\n";
  return unless $topicType =~ /\bEventTopic\b/;

  my $uid = $meta->getPath();
  $uid =~ s/\//./g;
  my $web = $meta->web;
  my $topic = $meta->topic;

  return $this->createEvent(
    uid => $uid,
    location => Foswiki::Func::getScriptUrl($web, $topic, 'view'),
    summary => Foswiki::Plugins::Func::getTopicTitle($web, $topic, undef, $meta),
    description => $this->getField($meta, "Summary"),
    start => $this->getField($meta, 'StartDate'),
    end => $this->getField($meta, 'EndDate'),
    freq => $this->getField($meta, 'Recurrence') || 'none',
    weekday => $this->getField($meta, 'WeekDay'),
    montday => $this->getField($meta, 'MonthDay'),
    month => $this->getField($meta, 'Month'),
    yearday => $this->getField($meta, 'YearDay'),
    interval => $this->getField($meta, 'Interval'),
    #count => $this->getField($meta, 'RecurrenceCount'), # unusede for now
  );
}

###############################################################################
sub getEventFromMetaData {
  my ($this, $web, $topic, $record) = @_;

  $web =~ s/\//./g;
  my $uid = $web . '.' . $topic . '#' . $record->{name};

  my $event = $this->createEvent(

    uid => $uid,
    location => Foswiki::Func::getScriptUrl($web, $topic, 'view'),
    summary => $record->{TopicTitle} || '',
    description => $record->{Summary} || '',
    start => $record->{StartDate},
    end => $record->{EndDate},
    freq => $record->{Recurrence} || 'none',
    weekday => $record->{WeekDay},
    monthday => $record->{MonthDay},
    month => $record->{Month},
    yearday => $record->{YearDay},
    interval => $record->{Interval},

    # count => $record->{count}, # unused for now
  );
}

###############################################################################
sub getEventsFromMetaData {
  my ($this, $meta) = @_;

  my $web = $meta->web;
  my $topic = $meta->topic;

  my @events = ();

  # SMELL: there's no way to iterate over all known meta data keys
  # so we have to break data encapsulation of Foswiki::Meta
  while (my ($alias, $name) = each %Foswiki::Meta::aliases) {
    $name =~ s/^META://;

    # SMELL: there's no way to access the meta data definition record for
    # a given key
    my $metaDataDef = $Foswiki::Meta::VALIDATE{$name};
    next unless $metaDataDef;

    # SMELL: there's a form attribute stored into the record by MetaDataPlugin,
    # that's not standard, yet we need it now to read the TopicType formfield
    my $formWebTopic = $metaDataDef->{form};
    next unless $formWebTopic;

    # now, we've found a meta data definition that has got a DataForm associated with it
    my ($formWeb, $formTopic) = Foswiki::Func::normalizeWebTopicName($web, $formWebTopic);
    next unless Foswiki::Func::topicExists($formWeb, $formTopic);

    # now check the TopicType for "EventTopic"
    my $form;
    try {
      $form = new Foswiki::Form($Foswiki::Plugins::SESSION, $formWeb, $formTopic);
    }
    catch Foswiki::AccessControlException with {
      # catch but simply bail out
      _writeDebug("access exception reading $formWeb.$formTopic");
    };
    next unless $form;

    my $topicTypeField = $form->getField('TopicType');
    next unless $topicTypeField;

    my $topicType = $topicTypeField->{value};
    next unless $topicType || $topicType !~ /\bEventTopic\b/;

    # now we've found an EventTopic, we can extract the metadata for this key
    foreach my $record ($meta->find($name)) {
      push @events, $this->getEventFromMetaData($web, $topic, $record);
    }
  }

  return @events;
}

###############################################################################
sub createEvent {
  my ($this, %params) = @_;

  # create new event entry
  my $vevent = Data::ICal::Entry::Event->new();
  $vevent->add_properties(
    uid => $params{uid},
    summary => $params{summary},
    description => $params{description},
  );

  # datetime properties
  my %event = ();

  if (defined $params{start} && $params{start} ne '') {
    $event{dtstart} = DateTime->from_epoch(epoch => parseTime($params{start}));

    my $icalStart = DateTime::Format::ICal->format_datetime($event{dtstart});
    $icalStart =~ s/T.*//;    # strip of time
    $vevent->add_properties('dtstart' => [$icalStart, {VALUE => 'DATE'}]);
  }

  if ($params{end}) {
    $event{dtend} = DateTime->from_epoch(epoch => parseTime($params{end}));
    my $icalEnd = DateTime::Format::ICal->format_datetime($event{dtend});
    $icalEnd =~ s/T.*//;      # strip of time
    $vevent->add_properties('dtend' => [$icalEnd, {VALUE => 'DATE'}]);
  }

  # other datetime properties
  foreach my $prop (qw(created dtstamp last-modified)) {
    my $val = $params{$prop};
    next unless defined $val;
    $val = DateTime->from_epoch(epoch => parseTime($params{$prop}));
    $vevent->add_properties($prop => $val);
  }

  # optional unique
  foreach my $prop (qw(location organizer priority recurrence-id sequence status transp uid url)) {
    my $val = $params{$prop};
    $vevent->add_properties($prop => $val) if defined $val;
  }

  # optional repeatable
  # - attach
  # - attendee
  # - categories
  # - comment
  # - contact
  # - exdate
  # - exrule
  # - request-status
  # - related-to
  # - resources

  if (defined $params{freq} && $params{freq} ne 'none') {
    $event{freq} = $params{freq};
    $event{interval} = $params{interval} if $params{interval};
    $event{until} = $event{dtend} if defined $event{dtend};
    #$event{count} = $params{count}; # unused for now

    if ($params{freq} eq 'daily') {
      # nothing else
    } elsif ($params{freq} eq 'weekly') {
      $event{byday} = $params{weekday};
    } elsif ($params{freq} eq 'monthly') {
      $event{bymonthday} = $params{monthday};
    } elsif ($params{freq} eq 'yearly') {
      $event{bymonth} = $params{month} if defined $params{month};
      $event{bymonthday} = $params{monthday} if defined $params{monthday};
      $event{byyearday} = $params{yearday} if defined $params{yearday};
    }

    $event{byday} = [split(/\s*,\s*/, $event{byday})] if defined $event{byday};
    $event{bymonth} = [split(/\s*,\s*/, $event{bymonth})] if defined $event{bymonth};
    $event{bymonthday} = [split(/\s*,\s*/, $event{bymonthday})] if defined $event{bymonthday};
    $event{byyearday} = [split(/\s*,\s*/, $event{byyearday})] if defined $event{byyearday};

    my $set;
    my $error;

    try {
      $set = DateTime::Event::ICal->recur(%event);    # needs DateTimes
    }
    catch Error::Simple with {
      $error = shift;
      print STDERR "ERROR: " . $error . "\n";
    };

    unless ($error) {
      my @recurr = DateTime::Format::ICal->format_recurrence($set);
      foreach my $recurr (@recurr) {
        #print STDERR "recurr=$recurr\n";
        if ($recurr =~ s/^RRULE://) {
          $vevent->add_properties(rrule => $recurr);
        } elsif ($recurr =~ s/^RDATE://) {
          foreach my $rdate (split(/,/, $recurr)) {
            $vevent->add_properties(rdate => [$rdate, {VALUE => 'DATE'}]);
          }
        }
      }
    }
  }

  _writeDebug("generated vevent " . $vevent->as_string());
  return $vevent;
}

###############################################################################
sub getField {
  my ($this, $meta, $name) = @_;

  return unless $meta;

  my $field = $meta->get('FIELD', $name);
  return unless $field;

  return $field->{value};
}

###############################################################################
sub _inlineError {
  return '<span class="foswikiAlert">' . $_[0] . '</span>';
}

###############################################################################
sub _writeDebug {
  my $msg = shift;
  chomp($msg);
  print STDERR "ICalPlugin::Core - $msg\n" if TRACE;
}

1;
